# coop_config/main.py

"""
This is the entry point of the library in which the main config class is defined
"""


import json
import os
from .exceptions import ConfigKeyError, SectionError



class ConfigSection:
	"""
	this class represents config sections and its nested sections. It is intended to be used dynamically through the Config class itself except for the add_setting and add_section methods.

	"""

	def __init__(self, root, name, settings):
		self.root = root
		self.name = name
		self.settings = settings
		for key, value in self.settings.items():
			if isinstance(value, dict):
				section = ConfigSection(self.root, key, value)
				self._create_section_property(key, section)
				continue
			self._create_property(key)
	def _create_property(self, key):
		def getter(obj):
			return obj.settings.get(key)
		def setter(obj, value):
			self.settings[key] = value
			self.root.save_settings()
		setattr(self.__class__, key, property(getter, setter))
	
	def _create_section_property(self,key, section):
		def getter(obj):
			return section
		setattr(self.__class__, key, property(getter))
	def add_setting(self, key, value):
		"""
		function add_setting
		adds a new config item to the section
		parameters:
		key: the name of the new setting
		value: the default value of the new setting
		
		"""
		if isinstance(value, dict):
			raise ConfigKeyError("cannot use add_setting method to add dictionaries")
		self.settings[key] = value
		self._create_property(key)
		self.root.save_settings()
	def add_section(self, name, settings={}):
		"""function add_section
		adds a nested section to the current section
		parameters:
		name:str: the name of the section
		settings: dict: the initial config items contained in this section

		"""
		if not isinstance(settings, dict):
			raise SectionError("settings parameter only accepts dict type")
		self.settings[name] = settings
		self._create_section_property(name, ConfigSection(self.root, name, settings))
		self.root.save_settings()

	def __repr__(self):
		return f"ConfigSection: {self.name}\n{self.settings}"


class Config:
	"""
	this is the main class where object files are managed and created
	parameters of the constructor:
	path: the path of the json file
	defaults: dict: default values of the settings file
	"""

	def __init__(self, settings_path, defaults={}):
		self.defaults = defaults
		self.settings_path = settings_path
		self._load_settings() if os.path.exists(self.settings_path) else self._setup_file()
		for key, value in self.settings.items():
			if isinstance(value, dict):
				self._create_section_property(key, ConfigSection(self, key, value))
				continue
			self._create_property(key)
	def _create_section_property(self,key, section):
		def getter(obj):
			return section
		setattr(self.__class__, key, property(getter))

	def _load_settings(self):
		with open(self.settings_path, encoding='utf-8') as file:
			self.settings = json.load(file)

	def _setup_file(self):
		with open(self.settings_path, "w", encoding="utf-8") as file:
			json.dump(self.defaults, file, ensure_ascii=False, indent=4)
		self.settings = self.defaults
	def save_settings(self):
		with open(self.settings_path, "w", encoding="utf-8") as file:
			json.dump(self.settings, file, ensure_ascii=False, indent=4)

	def _create_property(self, key):
		def getter(obj):
			try:
				value = self.settings[key]
			except KeyError:
				value = self.defaults[key]
				self.settings[key] = value
				self.save_settings()
			return value
		def setter(obj, value):
			self.settings[key] = value
			self.save_settings()
		setattr(self.__class__, key, property(getter, setter))
	def add_setting(self, key, value):
		"""
		a function to add a new setting to the main config object based on the given name and value
		"""
		if isinstance(value, dict):
			raise ConfigKeyError("cannot use add_setting method to add dictionaries")
		self.settings[key] = value
		self._create_property(key)
		self.save_settings()
	def add_section(self, name, settings={}):
		"""
		a function to add a new section to the main config object based on the given name and initial settings
		"""
		if not isinstance(settings, dict):
			raise SectionError("settings parameter only accepts dict type")
		self.settings[name] = settings
		self._create_section_property(name, ConfigSection(self, name, settings))
		self.save_settings()
